import React from 'react'
import { Graph, Addon, Path, Cell, Model, Node } from '@antv/x6'
import { Drawer } from 'antd'
import AlgoNode from './components/AlgoNode'
import ToolBar from './components/ToolBar'
import DatasourceNodeInfo from './components/DatasourceNodeInfo'
import ScriptNodeInfo from './components/ScriptNodeInfo'
import './app.css'

// 侧边栏UI组件
const { Stencil } = Addon

// 节点状态接口
interface NodeStatus {
  id: string
  status: 'default' | 'success' | 'failed' | 'running'
  type: 'script' | 'datasource'
  label?: string
}

// 数据源节点
const dataSourceNodes = [
  {
    "id": "0",
    "shape": "dag-node",
    "height": 80,
    "width": 180,
    "data": {
      "label": "DataSource",
      "status": "default",
      "type": 'datasource'
    },
    "ports": {
      "items": [
        {"id": 'bottomPort', "group": 'bottom'},
      ]
    }
  }
]

// 数据处理节点
const dataHandleNodes = [
  {
    "id": "1",
    "shape": "dag-node",
    "data": {
      "label": "数据处理节点",
      "status": "default",
      "type": 'script'
    },
    "ports": {
      "items": [
        {"id": 'topPort', "group": 'top'},
        {"id": 'bottomPort', "group": 'bottom'},
      ]
    }
  }
]

// 其它节点
const otherScriptNodes = [
  {
    "id": "2",
    "shape": "dag-node",
    "data": {
      "label": "其它节点",
      "status": "default",
      "type": 'script'
    },
    "ports": {
      "items": [
        {"id": 'topPort', "group": 'top'},
        {"id": 'bottomPort', "group": 'bottom'},
      ]
    }
  },
]

// 用户自定义脚本节点
const customScriptNodes = [
  {
    "id": "5",
    "shape": "dag-node",
    "data": {
      "label": "SQL",
      "status": "default",
      "type": 'script'
    },
    "ports": {
      "items": [
        {"id": 'topPort', "group": 'top'},
        {"id": 'bottomPort', "group": 'bottom'},
      ]
    }
  },
  {
    "id": "6",
    "shape": "dag-node",
    "data": {
      "label": "PySpark",
      "status": "default",
      "type": 'script'
    },
    "ports": {
      "items": [
        {"id": 'topPort', "group": 'top'},
        {"id": 'bottomPort', "group": 'bottom'},
      ]
    }
  },
  {
    "id": "7",
    "shape": "dag-node",
    "data": {
      "label": "Python",
      "status": "default",
      "type": 'script'
    },
    "ports": {
      "items": [
        {"id": 'topPort', "group": 'top'},
        {"id": 'bottomPort', "group": 'bottom'},
      ]
    }
  },
]

// 注册算子节点
Graph.registerNode(
  'dag-node',
  {
    inherit: 'react-shape',
    width: 180,
    height: 36,
    component: <AlgoNode />,
    ports: {
      groups: {
        top: {
          position: 'top',
          attrs: {
            circle: {
              r: 4,
              magnet: true,
              stroke: '#C2C8D5',
              strokeWidth: 1,
              fill: '#fff',
            },
          },
        },
        bottom: {
          position: 'bottom',
          attrs: {
            circle: {
              r: 4,
              magnet: true,
              stroke: '#C2C8D5',
              strokeWidth: 1,
              fill: '#fff',
            },
          },
        },
      },
    },
  },
  true,
)

// 注册边
Graph.registerEdge(
  'dag-edge',
  {
    inherit: 'edge',
    attrs: {
      line: {
        stroke: '#C2C8D5',
        strokeWidth: 1,
        targetMarker: 'block',
      },
    },
  },
  true,
)

// 注册连接
Graph.registerConnector(
  'algo-connector',
  (s, e) => {
    const offset = 4
    const deltaY = Math.abs(e.y - s.y)
    const control = Math.floor((deltaY / 3) * 2)

    const v1 = { x: s.x, y: s.y + offset + control }
    const v2 = { x: e.x, y: e.y - offset - control }

    return Path.normalize(
      `M ${s.x} ${s.y}
       L ${s.x} ${s.y + offset}
       C ${v1.x} ${v1.y} ${v2.x} ${v2.y} ${e.x} ${e.y - offset}
       L ${e.x} ${e.y}
      `,
    )
  },
  true,
)

// 存储边
const dagEdges: Array<any> = []
export default class Dag extends React.Component<any, any> {
  // 容器DIV
  private container: HTMLDivElement
  // 侧边栏UI容器
  private stencilContainer: HTMLDivElement
  // 构造函数
  constructor(props: any) {
    super(props);
    this.state = {
      nodeInfoVisible: false,
      dagGraph: null,
      selectedNodeData: {}
    };
  }
  
  // 展示node节点信息
  showNodeInfo = () => {
    this.setState({nodeInfoVisible: true})
  }

  // 隐藏node节点信息
  hideNodeInfo = () => {
    this.setState({nodeInfoVisible: false})
  }


  // 挂载
  componentDidMount(): void {
    // 创建图
    const graph: Graph = new Graph({
      // 容器
      container: this.container,
      // 格子
      // grid: true,
      background: {
        color: '#FBFCFF'
      },
      // 拖动
      panning: {
        enabled: true,
        eventTypes: ['leftMouseDown', 'mouseWheel'],
      },
      // 放大缩小
      mousewheel: {
        enabled: true,
        modifiers: 'ctrl',
        factor: 1.1,
        maxScale: 1.5,
        minScale: 0.5,
      },
      // 交互触发高亮
      highlighting: {
        magnetAdsorbed: {
          name: 'stroke',
          args: {
            attrs: {
              fill: '#fff',
              stroke: '#31d0c6',
              strokeWidth: 4,
            },
          },
        },
      },
      // 连接
      connecting: {
        snap: true,
        allowBlank: false,
        allowLoop: false,
        highlight: true,
        connector: 'algo-connector',
        connectionPoint: 'anchor',
        anchor: 'center',
        // 验证
        validateMagnet({ magnet }) {          
          return magnet.getAttribute('port-group') !== 'top'
        },
        validateConnection({targetPort, sourceCell, targetCell}) {
          const jugeEdge = dagEdges.find((item: any) => (item?.source === sourceCell?.id) && (item?.target === targetCell?.id))
          if (jugeEdge) {
            return false
          }
          return targetPort === 'topPort'
        },
        createEdge() {
          return graph.createEdge({
            shape: 'dag-edge',
            attrs: {
              line: {
                strokeDasharray: '5 5',
              },
            },
            zIndex: -1,
          })
        },
      },
      // 选择
      selecting: {
        enabled: true,
        multiple: true,
        rubberEdge: true,
        rubberNode: true,
        modifiers: 'shift',
        rubberband: true,
      },
      // 快捷键
      keyboard: {
        enabled: true,
        global: true,
      },
    })

    // 删除节点
    graph.bindKey(['delete', 'backspace'], () => {
      graph.getSelectedCells().forEach((item: any) => {
        item.remove()
      })
    })

    // 监听边连接
    graph.on('edge:connected', ({ edge }) => {
      
      dagEdges.push({source: edge.getSourceCell()?.id, target: edge.getTargetCell()?.id})
      const targetNodeData = edge.getTargetCell()?.data
      targetNodeData.inputNumber += 1
      edge.attr({
        line: {
          strokeDasharray: '',
        },
      })
    })

    graph.on('edge:removed', ({edge}) => {
      const sourceNodeId = (edge as any).store.data.source.cell
      const targetNodeId = (edge as any).store.data.target.cell
      
      const index = dagEdges.indexOf({
        source: sourceNodeId,
        target: targetNodeId
      })

      const targetNode = this.state.dagGraph.getCellById(targetNodeId)
      
      if (targetNode) {
        targetNode.data.inputNumber -= 1
      }

      dagEdges.splice(index, 1)

    })

    // 监听状态改变
    graph.on('node:change:data', ({ node }) => {
      const edges = graph.getIncomingEdges(node)
      const { status } = node.getData() as NodeStatus
      edges?.forEach((edge) => {
        if (status === 'running') {
          edge.attr('line/strokeDasharray', 5)
          edge.attr('line/style/animation', 'running-line 30s infinite linear')
        } else {
          edge.attr('line/strokeDasharray', '')
          edge.attr('line/style/animation', '')
        }
      })
    })

    // 监听节点被双击
    graph.on('node:dblclick', (args: { 
      cell: Cell
      node: Node 
      options: Model.SetOptions 
    }) => {
      const nodeData = args.cell.data
      this.setState({selectedNodeData: nodeData})
      this.showNodeInfo()
    })

    // 内容居中
    graph.centerContent()

    // 创建侧边栏
    const stencil = new Stencil({
      title: '算子检索',
      target: graph,
      // 搜索
      search(cell, keyword) {
        return cell.data.label.indexOf(keyword) !== -1
      },
      placeholder: '搜索算子',
      notFoundText: '未找到相关算子',
      scaled: true,
      // animation: true,
      // 侧边栏项宽高
      stencilGraphWidth: 200,
      stencilGraphHeight: 200,
      layoutOptions: {
        columns: 1,
        dx: 55,
        rowHeight: 'compact'
      },
      // 侧边栏项目分组
      groups: [
        {
          name: 'dataSource',
          title: '数据源'
        },
        {
          name: 'dataHandle',
          title: '数据处理',
          collapsed: true
        },
        {
          name: 'otherScript',
          title: '其它',
          collapsed: true
        },
        {
          name: 'customScript',
          title: '自定义脚本'
        }
      ],
      // 拖拽事件 处理节点数据
      getDropNode(draggingNode: Node): Node {
        const newNode = draggingNode.clone()
        initNodeData(newNode)
        return newNode
      },
    })

    // 添加到侧边栏容器内
    this.stencilContainer.appendChild(stencil.container)

    // 将可拖拽项加载到侧边栏中
    stencil.load(dataSourceNodes, 'dataSource')
    stencil.load(dataHandleNodes, 'dataHandle')
    stencil.load(otherScriptNodes, 'otherScript')
    stencil.load(customScriptNodes, 'customScript')
    stencil.resizeGroup('dataSource', { width: 200, height: dataSourceNodes.length * 120 })
    stencil.resizeGroup('dataHandle', { width: 200, height: dataHandleNodes.length * 60 })
    stencil.resizeGroup('otherScript', { width: 200, height: otherScriptNodes.length * 60 })
    stencil.resizeGroup('customScript', { width: 200, height: customScriptNodes.length * 60 })    
    
    // 设置图
    this.setState({dagGraph: graph})
  }

  // 卸载
  componentWillUnmount(): void {
    this.state.dagGraph.dispose()
  }
  // 容器
  refContainer = (container: HTMLDivElement) => {
    this.container = container
  }

  // 侧边栏
  refStencil = (container: HTMLDivElement) => {
    this.stencilContainer = container
  }

  // 最后输出模板
  render() {
    return (
      <div className="app">
        {/* 侧边栏 */}
        <div className="app-stencil" ref={this.refStencil} />
        {/* 容器 */}
        <div className='app-content'>
          <div className="app-graph" ref={this.refContainer} />
          {/* 节点信息 */}
          <Drawer
            title={this.state.selectedNodeData?.label || 'NodeTitle'}
            placement="right"
            closable={false}
            destroyOnClose={true}
            onClose={this.hideNodeInfo}
            visible={this.state.nodeInfoVisible}
            getContainer={false}
            style={{ position: 'absolute' }}
          >
            {/* 数据源节点 */}
            {this.state.selectedNodeData?.type === 'datasource' && <DatasourceNodeInfo nodeInfo={this.state.selectedNodeData}/>}
            {/* 脚本节点 */}
            {this.state.selectedNodeData?.type === 'script' && <ScriptNodeInfo nodeInfo={this.state.selectedNodeData}/>}
          </Drawer>
          {/* 工具栏 */}
          <ToolBar className="app-tool" graph={this.state.dagGraph}/>
        </div>
      </div>
    )
  }
}

// 初始化节点数据
function initNodeData(newNode: Node<Node.Properties>) {
  // 节点id
  newNode.data.id = newNode.id
  switch(newNode.data.type) {
    case 'datasource':
      newNode.data.inputSource = undefined
      newNode.data.dataTable = undefined
      break
    case 'script':
      newNode.data.paramText = undefined
      newNode.data.scriptText = undefined
      newNode.data.outputData = undefined
      newNode.data.inputNumber = 0
      newNode.data.packageData = undefined
      break
    default:
  }
}

